﻿#region License Information

/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */

#endregion

#region

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;

#endregion

namespace HeuristicLab.Optimization {
  /// <summary>
  /// The generational Distance is defined as the pth-root of the sum of all d[i]^(p) divided by the size of the qualities
  ///  where d[i] is the minimal distance the ith point of the evaluated qualities has to any point in the optimal pareto qualities.   
  ///  p is a dampening factor and is normally set to 1. 
  ///  http://shodhganga.inflibnet.ac.in/bitstream/10603/15070/28/28_appendix_h.pdf
  /// </summary>
  public static class GenerationalDistanceCalculator {
    public static double CalculateGenerationalDistance<TP1, TP2>(IEnumerable<TP1> qualities, IEnumerable<TP2> bestKnownFront, double p) where TP1 : IReadOnlyList<double> where TP2 : IReadOnlyList<double> {
      if (qualities == null || bestKnownFront == null) throw new ArgumentNullException("Fronts must not be null");
      if (p.IsAlmost(0.0)) throw new ArgumentException("p must not be zero.");
      var mat = bestKnownFront.ToMatrix();
      if (mat.GetLength(0) == 0) throw new ArgumentException("Fronts must not be empty.");

      alglib.kdtree tree;
      alglib.kdtreebuild(mat, mat.GetLength(0), mat.GetLength(1), 0, 2, out tree);
      var sum = 0.0;
      var summand = new double[1];
      var count = 0;
      foreach(var point in qualities) {
        alglib.kdtreequeryknn(tree, point.ToArray(), 1, true);
        alglib.kdtreequeryresultsdistances(tree, ref summand);
        sum += Math.Pow(summand[0], p);
        count++;
      }

      if (count == 0) throw new ArgumentException("Fronts must not be empty.");
      return Math.Pow(sum, 1 / p) / count;
    }

    public static double CalculateInverseGenerationalDistance<TP1, TP2>(IEnumerable<TP1> qualities, IEnumerable<TP2> bestKnownFront, double p) where TP1 : IReadOnlyList<double> where TP2 : IReadOnlyList<double> {
      return CalculateGenerationalDistance(bestKnownFront, qualities, p);
    }

    internal static double[,] ToMatrix<TP>(this IEnumerable<TP> source) where TP : IReadOnlyList<double> {
      var l = source.ToArray();
      var firstDimension = l.Length;
      var secondDimension = l[0].Count;
      var result = new double[firstDimension, secondDimension];
      for (var i = 0; i < firstDimension; ++i)
        for (var j = 0; j < secondDimension; ++j)
          result[i, j] = l[i][j];

      return result;
    }
  }
}